from faststream_compressors.middlewares import BaseDecompressionMiddleware
from nats.aio.msg import Msg


class NatsDecompressionMiddleware(BaseDecompressionMiddleware):
    msg: Msg

    @property
    def content_type(self) -> str | None:
        if not self.msg.headers:
            return
        return self.msg.headers.get("content-type")

    @property
    def content_encoding(self) -> list[str] | None:
        if not self.msg.headers:
            return
        if "content-encoding" not in self.msg.headers:
            return
        return list(map(str.strip, self.msg.headers["content-encoding"].split(",")))

    @property
    def body(self) -> bytes:
        return self.msg.data

    @body.setter
    def body(self, value: bytes):
        self.msg.data = value


__all__ = ("NatsDecompressionMiddleware",)



from abc import abstractmethod
from functools import partial
from typing import Any, Callable, Sequence

from faststream import BaseMiddleware
from faststream.log.logging import logger
from faststream.broker.message import encode_message
from faststream.types import AsyncFunc

from faststream_compressors.compressors import BaseCompressor
from faststream_compressors.errors import UnknownEncoding
from faststream_compressors.handler import CompressorHandler, DecompressorHandler


class BaseDecompressionMiddleware(BaseMiddleware):
    """
    A base class for creating middleware to handle message decompression.

    This middleware is used to decompress data upon receiving messages.
    """

    def __init__(self, msg: Any, *, decompressors: Sequence[BaseCompressor]):
        self.decompressor_handler = DecompressorHandler(*decompressors)

        super().__init__(msg)

    @classmethod
    def make_middleware(
        cls, decompressors: Sequence[BaseCompressor] | BaseCompressor
    ) -> Callable[[Any], "BaseDecompressionMiddleware"]:
        """
        Creates a partial function that can be used to instantiate the middleware.

        :param decompressors: A sequence of decompressors or a single decompressor to use for message decompression.
        :return: A partial function to instantiate the middleware.
        """
        if isinstance(decompressors, BaseCompressor):
            decompressors = (decompressors,)
        return partial(cls, decompressors=decompressors)

    @property
    @abstractmethod
    def content_type(self) -> str | None:
        """
        Abstract property to get the content type of the message.

        :return: The content type of the message.
        """

    @property
    @abstractmethod
    def content_encoding(self) -> list[str] | None:
        """
        Abstract property to get the content encoding of the message.

        :return: A list of encodings applied to the message.
        """

    @property
    @abstractmethod
    def body(self) -> bytes: ...

    @body.setter
    @abstractmethod
    def body(self, value: bytes) -> None: ...

    async def on_receive(self) -> None:
        if not self.content_encoding:
            return

        try:
            self.body = self.decompressor_handler(
                self.body, *reversed(self.content_encoding)
            )
        except UnknownEncoding as exception:
            logger.warning(f"{self.__class__.__name__}: {exception}")
        except Exception as exception:
            logger.error("Failed to decompress message:", exc_info=exception)


class CompressionMiddleware(BaseMiddleware):
    """
    Middleware for compressing published messages
    """

    def __init__(self, msg: Any, *, compressors: Sequence[BaseCompressor]):
        self.compressor_handler = CompressorHandler(*compressors)

        super().__init__(msg)

    @classmethod
    def make_middleware(
        cls, compressors: Sequence[BaseCompressor] | BaseCompressor
    ) -> Callable[[Any], "CompressionMiddleware"]:
        """
        Creates a partial function that can be used to instantiate the middleware.

        :param compressors: A sequence of compressors or a single compressor to use for message compression.
        :return: A partial function to instantiate the middleware.
        """
        if isinstance(compressors, BaseCompressor):
            compressors = (compressors,)
        return partial(cls, compressors=compressors)

    async def publish_scope(
        self,
        call_next: "AsyncFunc",
        msg: Any,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if not self.compressor_handler.compressors:
            return await call_next(msg, *args, **kwargs)

        if not kwargs["headers"]:
            kwargs["headers"] = dict()

        if "content-encoding" in kwargs["headers"]:
            return await call_next(msg, *args, **kwargs)

        try:
            if "content-type" not in kwargs["headers"]:
                msg, content_type = encode_message(msg)
                if content_type:
                    kwargs["headers"]["content-type"] = content_type
            msg, kwargs["headers"]["content-encoding"] = self.compressor_handler(msg)
        except Exception as exception:
            logger.error("Failed to compress message:", exc_info=exception)

        return await call_next(msg, *args, **kwargs)


__all__ = ("BaseDecompressionMiddleware", "CompressionMiddleware")



from abc import ABC

from faststream_compressors.compressors import BaseCompressor
from faststream_compressors.errors import UnknownEncoding


class BaseCompressorHandler(ABC):
    """
    Abstract base class for compressor handlers.

    :param compressors: A sequence of compressors to be managed by the handler.
    """

    def __init__(self, *compressors: BaseCompressor):
        self.compressors = compressors

    def get_compressor(self, encoding: str) -> BaseCompressor:
        """
        Retrieves the compressor corresponding to the given encoding.

        :param encoding: The encoding of the compressor to retrieve.
        :returns: The compressor corresponding to the given encoding.
        :raises UnknownEncoding: If no compressor with the given encoding is found.
        """
        compressor: BaseCompressor | None = next(
            filter(lambda x: encoding == x.ENCODING, self.compressors), None
        )
        if not compressor:
            raise UnknownEncoding(encoding)
        return compressor


class CompressorHandler(BaseCompressorHandler):
    """Handles the compression of message data."""

    def __call__(self, data: bytes) -> tuple[bytes, str | None]:
        """
        Compresses the provided data using the sequence of compressors.

        :param data: The data to be compressed.
        :return: A tuple of the compressed data and a comma-separated content-type header value.
        """
        encodings = []
        for compressor in self.compressors:
            encodings.append(compressor.ENCODING)
            data = compressor(data)
        return data, ", ".join(encodings) if encodings else None


class DecompressorHandler(BaseCompressorHandler):
    """Handles the decompression of message data."""

    def __call__(self, data: bytes, *encodings: str) -> bytes:
        """
        Decompresses the provided data using the sequence of encodings.

        :param data: The data to be decompressed.
        :param encodings: A sequence of encodings to be used for decompression.
        :return: The decompressed data.
        """
        for encoding in encodings:
            compressor = self.get_compressor(encoding)
            data = compressor(data)
        return data


__all__ = ("CompressorHandler", "DecompressorHandler")



import gzip

from faststream_compressors.compressors import BaseCompressor


class GzipCompressor(BaseCompressor):
    """A class for compressing data using gzip."""

    ENCODING = "gzip"

    def __init__(self, compress_level: int = 9):
        """
        Initializes the GzipCompressor with the specified compression level.

        :param compress_level: The level of compression to use (1-9). Default is 9.
        """
        self.compress_level = compress_level

    def __call__(self, data: bytes) -> bytes:
        """
        Compresses the provided data using gzip.

        :param data: Data to be compressed.
        :returns: Gzip-compressed data.
        """
        return gzip.compress(data, compresslevel=self.compress_level)


class GzipDecompressor(BaseCompressor):
    """A class for decompressing gzip-compressed data."""

    ENCODING = "gzip"

    def __call__(self, data: bytes) -> bytes:
        """
        Decompresses the provided gzip-compressed data.

        :param data: Gzip-compressed data.
        :returns: Decompressed data.
        """
        return gzip.decompress(data)


__all__ = ("GzipCompressor", "GzipDecompressor")



from abc import ABC, abstractmethod
from typing import ClassVar


class BaseCompressor(ABC):
    """
    Abstract base class for all compressors.

    :cvar ENCODING: String representation of the compression scheme used. E.g. "gzip" and "lz4".
    """

    ENCODING: ClassVar[str]

    @abstractmethod
    def __call__(self, data: bytes) -> bytes:
        """
        Compresses the provided data.
        This method must be overridden in a subclass.

        :param data: Data to be compressed.
        :returns: Compressed data.
        """


from faststream_compressors.compressors.gzip import *  # noqa: F401

__all__ = ("BaseCompressor",) + gzip.__all__



class UnknownEncoding(ValueError):
    """Exception raised when an unknown encoding is encountered."""

    def __init__(self, encoding: str):
        super().__init__(f"Unknown encoding: {encoding!r}")


__all__ = ("UnknownEncoding",)



import lz4.frame

from faststream_compressors.compressors import BaseCompressor


class Lz4Compressor(BaseCompressor):
    """A class for compressing data using LZ4."""

    ENCODING = "lz4"

    def __init__(
        self,
        compression_level: int = 9,
        block_size: int = 0,
        content_checksum: bool = False,
        block_linked: bool = True,
        store_size: bool = False,
    ):
        """
        Initializes the Lz4Compressor.

        :param block_size: Sepcifies the maximum blocksize to use.
        :param  content_checksum: Specifies whether to enable checksumming
            of the uncompressed content. If True, a checksum is stored at the
            end of the frame, and checked during decompression.
        :param block_linked: Specifies whether to use block-linked
            compression. If ``True``, the compression ratio is improved,
            particularly for small block sizes.
        :param store_size: If ``True`` then the frame will include an 8-byte
            header field that is the uncompressed size of data included
            within the frame.
        """
        self.compression_level = compression_level
        self.block_size = block_size
        self.content_checksum = content_checksum
        self.block_linked = block_linked
        self.store_size = store_size

    def __call__(self, data: bytes) -> bytes:
        """
        Compresses the provided data using LZ4.

        :param data: Data to be compressed.
        :returns: LZ4-compressed data.
        """
        return lz4.frame.compress(
            data,
            compression_level=self.compression_level,
            block_size=self.block_size,
            content_checksum=self.content_checksum,
            block_linked=self.block_linked,
            store_size=self.store_size,
        )


class Lz4Decompressor(BaseCompressor):
    """A class for decompressing LZ4-compressed data."""

    ENCODING = "lz4"

    def __call__(self, data: bytes) -> bytes:
        """
        Decompresses the provided LZ4-compressed data.

        :param data: LZ4-compressed data.
        :returns: Decompressed data.
        """
        return lz4.frame.decompress(data)


__all__ = ("Lz4Compressor", "Lz4Decompressor")
